@@ -25,29 +25,33 @@
 # --------------------------------------------------------------------------
 
 import ast
+import logging
 import re
 import time
+import warnings
 try:
     from urlparse import urlparse, parse_qs
 except ImportError:
     from urllib.parse import urlparse, parse_qs
 
 import keyring
+import adal
 from oauthlib.oauth2 import BackendApplicationClient, LegacyApplicationClient
 from oauthlib.oauth2.rfc6749.errors import (
     InvalidGrantError,
     MismatchingStateError,
     OAuth2Error,
     TokenExpiredError)
-from requests import RequestException
+from requests import RequestException, ConnectionError
 import requests_oauthlib as oauth
 
-from msrest.authentication import OAuthTokenAuthentication
+from msrest.authentication import OAuthTokenAuthentication, Authentication
 from msrest.exceptions import TokenExpiredError as Expired
-from msrest.exceptions import (
-    AuthenticationError,
-    raise_with_traceback)
+from msrest.exceptions import AuthenticationError, raise_with_traceback
 
+from msrestazure.azure_cloud import AZURE_CHINA_CLOUD, AZURE_PUBLIC_CLOUD
+
+_LOGGER = logging.getLogger(__name__)
 
 def _build_url(uri, paths, scheme):
     """Combine URL parts.
@@ -100,13 +104,9 @@ class AADMixin(OAuthTokenAuthentication):
     - Token caching and retrieval
     - Default AAD configuration
     """
-    _auth_endpoint = "//login.microsoftonline.com"
-    _china_auth_endpoint = "//login.chinacloudapi.cn"
     _token_uri = "/oauth2/token"
     _auth_uri = "/oauth2/authorize"
     _tenant = "common"
-    _resource = 'https://management.core.windows.net/'
-    _china_resource = "https://management.core.chinacloudapi.cn/"
     _keyring = "AzureAAD"
     _case = re.compile('([a-z0-9])([A-Z])')
 
@@ -114,6 +114,7 @@ class AADMixin(OAuthTokenAuthentication):
         """Configure authentication endpoint.
 
         Optional kwargs may include:
+            - cloud_environment (msrestazure.azure_cloud.Cloud): A targeted cloud environment
             - china (bool): Configure auth for China-based service,
               default is 'False'.
             - tenant (str): Alternative tenant, default is 'common'.
@@ -123,13 +124,22 @@ class AADMixin(OAuthTokenAuthentication):
               is 'https://management.core.windows.net/'.
             - verify (bool): Verify secure connection, default is 'True'.
             - keyring (str): Name of local token cache, default is 'AzureAAD'.
+            - timeout (int): Timeout of the request in seconds.
+            - proxies (dict): Dictionary mapping protocol or protocol and 
+              hostname to the URL of the proxy.
         """
         if kwargs.get('china'):
-            auth_endpoint = self._china_auth_endpoint
-            resource = self._china_resource
+            err_msg = ("china parameter is deprecated, "
+                       "please use "
+                       "cloud_environment=msrestazure.azure_cloud.AZURE_CHINA_CLOUD")
+            warnings.warn(err_msg, DeprecationWarning)
+            self.cloud_environment = AZURE_CHINA_CLOUD
         else:
-            auth_endpoint = self._auth_endpoint
-            resource = self._resource
+            self.cloud_environment = AZURE_PUBLIC_CLOUD
+        self.cloud_environment = kwargs.get('cloud_environment', self.cloud_environment)
+
+        auth_endpoint = self.cloud_environment.endpoints.active_directory
+        resource = self.cloud_environment.endpoints.management
 
         tenant = kwargs.get('tenant', self._tenant)
         self.auth_uri = kwargs.get('auth_uri', _https(
@@ -139,9 +149,11 @@ class AADMixin(OAuthTokenAuthentication):
         self.verify = kwargs.get('verify', True)
         self.cred_store = kwargs.get('keyring', self._keyring)
         self.resource = kwargs.get('resource', resource)
+        self.proxies = kwargs.get('proxies')
+        self.timeout = kwargs.get('timeout')
         self.state = oauth.oauth2_session.generate_token()
         self.store_key = "{}_{}".format(
-            self._auth_endpoint.strip('/'), self.store_key)
+            auth_endpoint.strip('/'), self.store_key)
 
     def _check_state(self, response):
         """Validate state returned by AAD server.
@@ -184,7 +196,10 @@ class AADMixin(OAuthTokenAuthentication):
         :rtype: None
         """
         self.token = token
-        keyring.set_password(self.cred_store, self.store_key, str(token))
+        try:
+            keyring.set_password(self.cred_store, self.store_key, str(token))
+        except Exception as err:
+            _LOGGER.warning("Keyring cache token has failed: %s", str(err))
 
     def _retrieve_stored_token(self):
         """Retrieve stored token for new session.
@@ -254,6 +269,7 @@ class AADTokenCredentials(AADMixin):
     e.g. Python ADAL lib.
 
     Optional kwargs may include:
+    - cloud_environment (msrestazure.azure_cloud.Cloud): A targeted cloud environment
     - china (bool): Configure auth for China-based service,
       default is 'False'.
     - tenant (str): Alternative tenant, default is 'common'.
@@ -300,6 +316,7 @@ class UserPassCredentials(AADRefreshMixin, AADMixin):
     that 2-factor auth be disabled.
 
     Optional kwargs may include:
+    - cloud_environment (msrestazure.azure_cloud.Cloud): A targeted cloud environment
     - china (bool): Configure auth for China-based service,
       default is 'False'.
     - tenant (str): Alternative tenant, default is 'common'.
@@ -309,8 +326,11 @@ class UserPassCredentials(AADRefreshMixin, AADMixin):
       is 'https://management.core.windows.net/'.
     - verify (bool): Verify secure connection, default is 'True'.
     - keyring (str): Name of local token cache, default is 'AzureAAD'.
+    - timeout (int): Timeout of the request in seconds.
     - cached (bool): If true, will not attempt to collect a token,
       which can then be populated later from a cached token.
+    - proxies (dict): Dictionary mapping protocol or protocol and
+      hostname to the URL of the proxy.
 
     :param str username: Account username.
     :param str password: Account password.
@@ -356,21 +376,23 @@ class UserPassCredentials(AADRefreshMixin, AADMixin):
 
         :raises: AuthenticationError if credentials invalid, or call fails.
         """
-        session = self._setup_session()
-        optional = {}
-        if self.secret:
-            optional['client_secret'] = self.secret
-        try:
-            token = session.fetch_token(self.token_uri, client_id=self.id,
-                                        username=self.username,
-                                        password=self.password,
-                                        resource=self.resource,
-                                        verify=self.verify,
-                                        **optional)
-        except (RequestException, OAuth2Error, InvalidGrantError) as err:
-            raise_with_traceback(AuthenticationError, "", err)
+        with self._setup_session() as session:
+            optional = {}
+            if self.secret:
+                optional['client_secret'] = self.secret
+            try:
+                token = session.fetch_token(self.token_uri, client_id=self.id,
+                                            username=self.username,
+                                            password=self.password,
+                                            resource=self.resource,
+                                            verify=self.verify,
+                                            proxies=self.proxies,
+                                            timeout=self.timeout,
+                                            **optional)
+            except (RequestException, OAuth2Error, InvalidGrantError) as err:
+                raise_with_traceback(AuthenticationError, "", err)
 
-        self.token = token
+            self.token = token
 
 
 class ServicePrincipalCredentials(AADRefreshMixin, AADMixin):
@@ -378,6 +400,7 @@ class ServicePrincipalCredentials(AADRefreshMixin, AADMixin):
     Authenticates via a Client ID and Secret.
 
     Optional kwargs may include:
+    - cloud_environment (msrestazure.azure_cloud.Cloud): A targeted cloud environment
     - china (bool): Configure auth for China-based service,
       default is 'False'.
     - tenant (str): Alternative tenant, default is 'common'.
@@ -387,8 +410,11 @@ class ServicePrincipalCredentials(AADRefreshMixin, AADMixin):
       is 'https://management.core.windows.net/'.
     - verify (bool): Verify secure connection, default is 'True'.
     - keyring (str): Name of local token cache, default is 'AzureAAD'.
+    - timeout (int): Timeout of the request in seconds.
     - cached (bool): If true, will not attempt to collect a token,
       which can then be populated later from a cached token.
+    - proxies (dict): Dictionary mapping protocol or protocol and
+      hostname to the URL of the proxy.
 
     :param str client_id: Client ID.
     :param str secret: Client secret.
@@ -423,105 +449,101 @@ class ServicePrincipalCredentials(AADRefreshMixin, AADMixin):
 
         :raises: AuthenticationError if credentials invalid, or call fails.
         """
-        session = self._setup_session()
-        try:
-            token = session.fetch_token(self.token_uri, client_id=self.id,
-                                        resource=self.resource,
-                                        client_secret=self.secret,
-                                        response_type="client_credentials",
-                                        verify=self.verify)
-        except (RequestException, OAuth2Error, InvalidGrantError) as err:
-            raise_with_traceback(AuthenticationError, "", err)
-        else:
-            self.token = token
+        with self._setup_session() as session:
+            try:
+                token = session.fetch_token(self.token_uri, client_id=self.id,
+                                            resource=self.resource,
+                                            client_secret=self.secret,
+                                            response_type="client_credentials",
+                                            verify=self.verify,
+                                            timeout=self.timeout,
+                                            proxies=self.proxies)
+            except (RequestException, OAuth2Error, InvalidGrantError) as err:
+                raise_with_traceback(AuthenticationError, "", err)
+            else:
+                self.token = token
 
+# For backward compatibility of import, but I doubt someone uses that...
+class InteractiveCredentials(object):
+    def __init__(self, *args, **kwargs):
+        raise NotImplementedError("InteractiveCredentials was not functionning and was removed. Please use ADAL and device code instead.")
 
-class InteractiveCredentials(AADMixin):
-    """Credentials object for Interactive/Web App Authentication.
-    Requires that an AAD Client be configured with a redirect URL.
+class AdalAuthentication(Authentication):  # pylint: disable=too-few-public-methods
+    """A wrapper to use ADAL for Python easily to authenticate on Azure.
 
-    Optional kwargs may include:
-    - china (bool): Configure auth for China-based service,
-      default is 'False'.
-    - tenant (str): Alternative tenant, default is 'common'.
-    - auth_uri (str): Alternative authentication endpoint.
-    - token_uri (str): Alternative token retrieval endpoint.
-    - resource (str): Alternative authentication resource, default
-      is 'https://management.core.windows.net/'.
-    - verify (bool): Verify secure connection, default is 'True'.
-    - keyring (str): Name of local token cache, default is 'AzureAAD'.
-    - cached (bool): If true, will not attempt to collect a token,
-      which can then be populated later from a cached token.
-
-    :param str client_id: Client ID.
-    :param str redirect: Redirect URL.
+    .. versionadded:: 0.4.5
     """
 
-    def __init__(self, client_id, redirect, **kwargs):
-        super(InteractiveCredentials, self).__init__(client_id, None)
-        self._configure(**kwargs)
+    def __init__(self, adal_method, *args, **kwargs):
+        """Take an ADAL `acquire_token` method and its parameters.
 
-        self.redirect = redirect
-        if not kwargs.get('cached'):
-            self.set_token()
+        :Example:
 
-    @classmethod
-    def retrieve_session(cls, client_id, redirect):
-        """Create InteractiveCredentials from a cached token if it has not
-        yet expired.
+        .. code:: python
+
+            context = adal.AuthenticationContext('https://login.microsoftonline.com/ABCDEFGH-1234-1234-1234-ABCDEFGHIJKL')
+            RESOURCE = '00000002-0000-0000-c000-000000000000' #AAD graph resource
+            token = context.acquire_token_with_client_credentials(
+                RESOURCE,
+                "http://PythonSDK",
+                "Key-Configured-In-Portal")
+
+        can be written here:
+
+        .. code:: python
+
+            context = adal.AuthenticationContext('https://login.microsoftonline.com/ABCDEFGH-1234-1234-1234-ABCDEFGHIJKL')
+            RESOURCE = '00000002-0000-0000-c000-000000000000' #AAD graph resource
+            credentials = AdalAuthentication(
+                context.acquire_token_with_client_credentials,
+                RESOURCE,
+                "http://PythonSDK",
+                "Key-Configured-In-Portal")
+
+        or using a lambda if you prefer:
+
+        .. code:: python
+
+            context = adal.AuthenticationContext('https://login.microsoftonline.com/ABCDEFGH-1234-1234-1234-ABCDEFGHIJKL')
+            RESOURCE = '00000002-0000-0000-c000-000000000000' #AAD graph resource
+            credentials = AdalAuthentication(
+                lambda: context.acquire_token_with_client_credentials(
+                    RESOURCE,
+                    "http://PythonSDK",
+                    "Key-Configured-In-Portal"
+                )
+            )
+
+        :param adal_method: A lambda with no args, or `acquire_token` method with args using args/kwargs
+        :param args: Optional args for the method
+        :param kwargs: Optional kwargs for the method
         """
-        session = cls(client_id, redirect, cached=True)
-        session._retrieve_stored_token()
-        return session
+        self._adal_method = adal_method
+        self._args = args
+        self._kwargs = kwargs
 
-    def _setup_session(self):
-        """Create token-friendly Requests session.
+    def signed_session(self):
+        """Get a signed session for requests.
 
-        :rtype: requests_oauthlib.OAuth2Session
+        Usually called by the Azure SDKs for you to authenticate queries.
+
+        :rtype: requests.Session
         """
-        return oauth.OAuth2Session(self.id,
-                                   redirect_uri=self.redirect,
-                                   state=self.state)
+        session = super(AdalAuthentication, self).signed_session()
 
-    def get_auth_url(self, msa=False, **additional_args):
-        """Get URL to web portal for authentication.
-
-        :param bool msa: Set to 'True' if authenticating with Live ID. Default
-         is 'False'.
-        :param additional_args: Set and additional kwargs for requrired AAD
-         configuration: msdn.microsoft.com/en-us/library/azure/dn645542.aspx
-        :rtype: Tuple
-        :return: The URL for authentication (str), and state code that will
-         be verified in the response (str).
-        """
-        if msa:
-            additional_args['domain_hint'] = 'live.com'
-        session = self._setup_session()
-        auth_url, state = session.authorization_url(self.auth_uri,
-                                                    resource=self.resource,
-                                                    **additional_args)
-        return auth_url, state
-
-    def set_token(self, response_url):
-        """Get token using Authorization Code from redirected URL.
-
-        :param str response_url: The full redirected URL from successful
-         authentication.
-        :raises: AuthenticationError if credentials invalid, or call fails.
-        """
-        self._check_state(response_url)
-        session = self._setup_session()
-
-        if response_url.startswith(_http(self.redirect)):
-            response_url = _https(response_url)
-        elif not response_url.startswith(_https(self.redirect)):
-            response_url = _https(self.redirect, response_url)
         try:
-            token = session.fetch_token(self.token_uri,
-                                        authorization_response=response_url,
-                                        verify=self.verify)
-        except (InvalidGrantError, OAuth2Error,
-                MismatchingStateError, RequestException) as err:
-            raise_with_traceback(AuthenticationError, "", err)
-        else:
-            self.token = token
+            raw_token = self._adal_method(*self._args, **self._kwargs)
+        except adal.AdalError as err:
+            # pylint: disable=no-member
+            if (hasattr(err, 'error_response') and ('error_description' in err.error_response)
+                    and ('AADSTS70008:' in err.error_response['error_description'])):
+                raise Expired("Credentials have expired due to inactivity.")
+            else:
+                raise AuthenticationError(err)
+        except ConnectionError as err:
+            raise AuthenticationError('Please ensure you have network connection. Error detail: ' + str(err))
+
+        scheme, token = raw_token['tokenType'], raw_token['accessToken']
+        header = "{} {}".format(scheme, token)
+        session.headers['Authorization'] = header
+        return session
